//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_ProtocolNETWORKSERVICESERVER_H
#define PROTOCOL_ProtocolNETWORKSERVICESERVER_H

#include <thread>
#include "NetworkMessage.h"

#include "Protocol.h"
#include "Networking.h"

namespace networking
{
namespace service
{

template<class Service, class Protocol>
class Server
{
};

// TCP Implementation
template<class Service>
class Server<Service, boost::asio::ip::tcp>
{
public:
    using RequestMessage = typename Service::RequestMessage;
    using ResponseMessage = typename Service::ResponseMessage;

    using Endpoint = typename boost::asio::ip::tcp::endpoint;
    using Socket = typename boost::asio::ip::tcp::socket;
    using Acceptor = typename boost::asio::ip::tcp::acceptor;

    using OnRequestReceivedCallback =
    std::function<ResponseMessage(const Endpoint & clientEndpoint, const RequestMessage & requestMessage)>;

    using OnNetworkErrorCallback = std::function<void(const boost::system::system_error & error)>;

    Server(Networking & net)
        : net(net)
          , acceptor(net.getIoService())
    {}

    ~Server()
    {
        stop();
    }

    void advertiseService(const Endpoint & bindEndpoint,
                          OnRequestReceivedCallback onRequestReceivedCallback,
                          OnNetworkErrorCallback onNetworkErrorCallback = defaultOnNetworkErrorCallback)
    {
        if (running)
            stop();

        running = true;

        this->onRequestReceivedCallback = onRequestReceivedCallback;
        this->onNetworkErrorCallback = onNetworkErrorCallback;

        acceptor = Acceptor(net.getIoService(), bindEndpoint);
        accept();
    }

    void advertiseService(const std::string & host, uint16_t port,
                          OnRequestReceivedCallback onRequestReceivedCallback,
                          OnNetworkErrorCallback onNetworkErrorCallback = defaultOnNetworkErrorCallback)
    {
        advertiseService(Endpoint(
            boost::asio::ip::address::from_string(host), port),
                         onRequestReceivedCallback, onNetworkErrorCallback);
    }

    void stop()
    {
        running = false;
        acceptor.cancel();
    }

private:
    Networking & net;
    OnRequestReceivedCallback onRequestReceivedCallback;
    OnNetworkErrorCallback onNetworkErrorCallback;
    Acceptor acceptor;
    bool running{false};

    void accept()
    {
        auto socket = std::make_shared<Socket>(net.getIoService());
        acceptor.async_accept(*socket, [this, socket](boost::system::error_code acceptError)
        {
            networking::utils::SafeSocketCloser<Socket> closer(*socket);

            if (!running)
                return;

            if (acceptError)
            {
                onNetworkErrorCallback(acceptError);
                return;
            }

            try
            {
                auto request = message::receive<RequestMessage>(*socket, boost::posix_time::seconds(3));
                auto response = onRequestReceivedCallback(socket->remote_endpoint(), request);
                message::send(*socket, response, boost::posix_time::seconds(3));
            }
            catch (const boost::system::system_error & error)
            {
                onNetworkErrorCallback(error);
            }

            // The next accept event will be put on the event queue.
            accept();
        });
    }

    //
    static void defaultOnNetworkErrorCallback(const boost::system::system_error & error)
    {}
};

namespace tcp
{
template<class Service>
using Server = typename networking::service::Server<Service, boost::asio::ip::tcp>;
}

namespace udp
{
template<class Service>
using Server = typename networking::service::Server<Service, boost::asio::ip::udp>;
}

}
}

#endif //PROTOCOL_ProtocolNETWORKSERVICESERVER_H
