//
// Created by philipp on 02.12.17.
//

#ifndef PROTOCOL_NETWORKUTILS_H
#define PROTOCOL_NETWORKUTILS_H

#include <boost/asio/ip/tcp.hpp>
#include <boost/asio.hpp>

namespace networking
{
namespace utils
{

/**
 * Connect a socket with a timeout.
 * @tparam Socket
 * @tparam EndpointIterator
 * @param socket
 * @param endpointIterator
 * @param timeout
 */
template<typename Socket, typename EndpointIterator>
void connect(Socket & socket,
             const EndpointIterator & endpointIterator,
             const boost::posix_time::time_duration & timeout)
{
    auto & ioService = socket.get_io_service();

    boost::asio::deadline_timer deadline(ioService);
    deadline.expires_from_now(timeout);
    // This tells the io_service to call the deadline handler when it either expires or was canceled.
    deadline.async_wait(
        [&socket](const boost::system::error_code & error)
        {
            // Check if our job was canceled.
            if (error)
                return;

            // Timeout expired: close the socket.
            boost::system::error_code ignoredError;
            socket.close(ignoredError);
        });

    // A 'would_block' connectError is guaranteed to never occur on an asynchronous operation.
    boost::system::error_code connectError = boost::asio::error::would_block;

    // Run asynchronous connect.
    boost::asio::async_connect(socket, endpointIterator,
                               [&connectError](const boost::system::error_code & error,
                                               EndpointIterator iter)
                               { connectError = error; }); // update connectError variable

    ioService.reset();
    // We wait run the io_service until "something happens" with the socket.
    do ioService.run_one(); while (connectError == boost::asio::error::would_block);

    // Determine whether a connection was successfully established.
    // Even though our deadline handler might have run to close the socket, the connect operation
    // might have notionally succeeded!
    if (connectError || !socket.is_open())
    {
        throw boost::system::system_error(
            connectError ? connectError : boost::asio::error::operation_aborted);
    }
}

/**
 * Writes data to a socket with timeout.
 * @tparam SyncReadStream
 * @tparam MutableBufferSequence
 * @param stream
 * @param buffer
 * @param timeout
 */
template<typename SyncReadStream, typename ConstBufferSequence>
void write(SyncReadStream & stream,
           const ConstBufferSequence & buffer,
           const boost::posix_time::time_duration & timeout)
{
    auto & ioService = stream.get_io_service();

    boost::asio::deadline_timer deadline(ioService);
    deadline.expires_from_now(timeout);

    deadline.async_wait(
        [&stream](const boost::system::error_code & error)
        {
            if (error)
                return;

            boost::system::error_code ignoredError;
            stream.cancel(ignoredError);
        });

    boost::system::error_code writeError = boost::asio::error::would_block;
    boost::asio::async_write(stream, buffer,
                             [&writeError](const boost::system::error_code & error, size_t bytesWritten)
                             { writeError = error; });

    ioService.reset();
    do ioService.run_one(); while (writeError == boost::asio::error::would_block);

    if (writeError)
        throw boost::system::system_error(writeError);
};

/**
 * Reads data from socket with timeout.
 * @tparam SyncReadStream
 * @tparam MutableBufferSequence
 * @param stream
 * @param buffer
 * @param timeout
 */
template<typename SyncReadStream, typename MutableBufferSequence>
void read(SyncReadStream & stream,
          const MutableBufferSequence & buffer,
          const boost::posix_time::time_duration & timeout)
{
    auto & ioService = stream.get_io_service();

    boost::asio::deadline_timer deadline(ioService);
    deadline.expires_from_now(timeout);

    deadline.async_wait(
        [&stream](const boost::system::error_code & error)
        {
            if (error)
                return;

            boost::system::error_code ignoredError;
            stream.cancel(ignoredError);
        });

    boost::system::error_code readError = boost::asio::error::would_block;
    boost::asio::async_read(stream, buffer,
                            [&readError](const boost::system::error_code & error, size_t)
                            { readError = error; });

    ioService.reset();
    do ioService.run_one(); while (readError == boost::asio::error::would_block);

    if (readError)
        throw boost::system::system_error(readError);
};

template<typename Socket>
void closeSafely(Socket & socket)
{
    boost::system::error_code ignoredError;
    socket.shutdown(boost::asio::ip::tcp::socket::shutdown_both, ignoredError);
    socket.close(ignoredError);
}

template<typename Socket>
class SafeSocketCloser
{
public:
    SafeSocketCloser(Socket & socket) : socket(socket)
    {}

    ~SafeSocketCloser()
    { closeSafely(socket); }

    SafeSocketCloser(const SafeSocketCloser & other) = delete;

    SafeSocketCloser(SafeSocketCloser && other) = delete;

    SafeSocketCloser & operator=(const SafeSocketCloser & other) = delete;

    SafeSocketCloser & operator=(SafeSocketCloser && other) = delete;

private:
    Socket & socket;
};

}
}

#endif //PROTOCOL_NETWORKUTILS_H
